(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Qheader ;;
open Math ;;


type frustum = {
  up : plane_t ;
  down : plane_t ;
  left : plane_t ;
  right : plane_t
} ;;


(*
A Faster Overlap Test for a Plane and a Bounding Box
based on Gem 1.7 in Graphics Gems IV (by Ned Greene)
Kenny Hoff 07/08/96
http://wwwx.cs.unc.edu/~hoff/research/vfculler/boxvfc/boxvfc.html
Version modifiee pour viter les cascades de tests. Il suffit d'extraire
les affectations constantes de chaque cas.
*)
(* val bbox_inside_plane : int array -> int array -> Qheader.plane_t -> bool *)
let bbox_inside_plane smin smax plane =
  let x = float (if plane.normal.x >= 0.0 then smax.(0) else smin.(0)) in
  let y = float (if plane.normal.y >= 0.0 then smax.(1) else smin.(1)) in
  let z = float (if plane.normal.z >= 0.0 then smax.(2) else smin.(2)) in
  plane.normal.x *. x +.
  plane.normal.y *. y +.
  plane.normal.z *. z >= plane.dist
;;


(* val node_is_in_frustum : frustum -> Qheader.node_t -> bool *)
let node_is_in_frustum frustum node =
  let smin = node.box.smin in
  let smax = node.box.smax in
  bbox_inside_plane smin smax frustum.up &&
  bbox_inside_plane smin smax frustum.down &&
  bbox_inside_plane smin smax frustum.left &&
  bbox_inside_plane smin smax frustum.right
;;


(* val leaf_is_in_frustum : frustum -> Qheader.dleaf_t -> bool *)
let leaf_is_in_frustum frustum leaf =
  let smin = leaf.lbound.smin in
  let smax = leaf.lbound.smax in
  bbox_inside_plane smin smax frustum.up &&
  bbox_inside_plane smin smax frustum.down &&
  bbox_inside_plane smin smax frustum.left &&
  bbox_inside_plane smin smax frustum.right
;;


(* val transform_frustum : frustum -> Camera.camera -> frustum *)
let transform_frustum frustum camera =
 let transform_plane { normal = normal ; dist = dist ; ptype = ptype } =
   (* D'abord tourner la normale au plan *)
   let normal' = vec3_rotate normal camera.Camera.rotation in
   (* Puis modifier sa distance en fonction de la translation *)
   (* The (cartesian) equation of a plane is linear in the coordinates x *)
   (* Note:                                                              *)
   (* and y, that is, of the form ax+by+cz+d=0. The normal direction to  *)
   (* this plane is (a,b,c). The intersection of this plane with the     *)
   (* x-axis, or x-intercept, is x=-d/a ; the y-intercept is y=-d/b, and *)
   (* the z-intercept is z=-d/c. The plane is vertical (perpendicular to *)
   (* the xy-plane) if c=0 ; it is perpendicular to the x-axis if b=c=0  *)
   (* and likewise for the other coordinates.                            *)
   (* When a*a+b*b+c*c=1 and d<=0 in the equation ax+by+cz+d=0, the      *)
   (* equation is said to be in normal form. In this case d is the       *)
   (* distance of the plane to the origin, and (a,b,c) are the direction *)
   (* cosines of the normal.                                             *)
   (* To reduce an arbitrary equation ax+by+cz+d=0 to normal form,       *)
   (* divide by +-sqrt(a*a+b*b+c*c), where the sign of the radical is    *)
   (* chosen opposite the sign of d when d <> 0, the same as the sign of *)
   (* c when d=0 and c <> 0, and the same as the sign of b otherwise.    *)
   let dist' =  camera.Camera.position.x *. normal'.x +.
		camera.Camera.position.y *. normal'.y +.
		camera.Camera.position.z *. normal'.z in
   { normal = normal' ; dist = dist' ; ptype = ptype } in
 { up = transform_plane frustum.up ;
   down = transform_plane frustum.down ;
   left = transform_plane frustum.left ;
   right = transform_plane frustum.right }
;;

