(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Qheader ;;
open Matrix ;;


let pi = 4.0 *. atan 1.0 ;;
let rad x = x *. pi /. 180.0 ;;



(* ******************
Normalisation de vecteur. C'est--dire, fait en sorte que x^2 + y^2 + z^2 = 1
*)
(* ******************
val vec3_normalize: Qheader.vec3_t -> Qheader.vec3_t
*)
let vec3_normalize { x = a1 ; y = a2 ; z = a3 } =
  let len = sqrt (a1 *. a1 +. a2 *. a2 +. a3 *. a3) in
  { x = a1 /. len ; y = a2 /. len ; z = a3 /. len }
;;



(* ******************
Rotation de vecteur suivant la rotation encode par la matrice 'matrix'
*)
(* ******************
val vec3_rotate: Qheader.vec3_t -> Matrix.matrix3x3 -> Qheader.vec3_t
*)
let vec3_rotate { x = x ; y = y ; z = z } matrix =
  let x' = x *. matrix.m11 +. y *. matrix.m12 +. z *. matrix.m13 in
  let y' = x *. matrix.m21 +. y *. matrix.m22 +. z *. matrix.m23 in
  let z' = x *. matrix.m31 +. y *. matrix.m32 +. z *. matrix.m33 in
  { x = x' ; y = y' ; z = z' }
;;



(* ******************
Calcule la distance d'un vertex  un plan. On parle ici de vertex dans le
monde 3D right-handed. Il ne s'agit pas du tout de vertex Glide.
Dans l'quation d'un plan, si le vecteur normal est normalis, alors le
coefficient d = distance du plan au point 0. C'est--dire que si le plan est
devant l'origine, la distance sera ngative.
*)
(* ******************
val vertex_plane_distance: Qheader.plane_t -> Gr.vertex -> float
*)
let vertex_plane_distance plane
    { Gr.base = { Gr.x = x ; Gr.y = y ; Gr.ooz = z } } =
  let num = plane.normal.x *. x +.
            plane.normal.y *. y +.
            plane.normal.z *. z +.
            plane.dist in
  (* Because normal vector of the plane is assumed to be normalized   *)
  (* we can avoid dividing by the numerator which is hence always 1.0 *)
  (* let denom = sqrt (plane.normal.x *. plane.normal.x +.
		    plane.normal.y *. plane.normal.y +.
		    plane.normal.z *. plane.normal.z) in *)
  num (* / denom *)
;;



type location = Front | Back ;;

(* ******************
Teste si un point est devant ou derrire un plan
*)
(* ******************
val point_plane_test: Qheader.plane_t -> Qheader.vec3_t -> location
*)
let point_plane_test plane { x = x ; y = y ; z = z } =
  (* Because normal vector of the plane is assumed to be *)
  (* normalized we can apply the same optimization than above. *)
  let tmp =
    plane.normal.x *. x +. plane.normal.y *. y +. plane.normal.z *. z in
    if tmp < plane.dist then Back else Front
;;



(* ******************
Projete en place un vertex. Attention, avant projection, le vertex est un
vertex du monde rel, c'est  dire dans un repre 3D right-handed. Ce n'est
donc pas rellement un vertex avec la smantique de Glide (pour qui les
vertices sont en fait toujours 2D). Aprs transformation, il devient un
vertex 2D purement  la Glide. Il y a donc un petit changement de smantique.
En particulier, x et y deviennent des coordonnes cran et non plus les
coordonnes dans le monde 3D qu'ils taient avant.
*)
(* ******************
val project_polygon3d_in_place: Gr.vertex array -> unit
*)
let project_polygon3d_in_place polygon =
  for i = 0 to Array.length polygon - 1 do
    let bvertex = polygon.(i).Gr.base in
    let tmuvertex = polygon.(i).Gr.tmu.(0) in
    let x = bvertex.Gr.x in
    let y = bvertex.Gr.y in
    let z = bvertex.Gr.ooz in
    (* Recentrer sur l'cran + compenser le fait que l'axe des x-cran   *)
    (* varient dans le sens inverse de l'axe des y dans le monde "rel". *)
    (* On multiplie par 320 = 640/2 pour le fov.                         *)
    bvertex.Gr.x <- 320.0 -. 320.0 *. y /. x ;
    bvertex.Gr.y <- 240.0 +. 320.0 *. z /. x ;
    bvertex.Gr.ooz <- x ;
    (* Partage de sous-expression commune *)
    let oow_factor = 1.0 /. x in
    bvertex.Gr.oow <- oow_factor ;
    (* Maintenant, les champs du tmu sont significatifs pour Glide *)
    tmuvertex.Gr.tmu_sow <- tmuvertex.Gr.tmu_sow *. oow_factor ;
    tmuvertex.Gr.tmu_tow <- tmuvertex.Gr.tmu_tow *. oow_factor
  done
;;
