(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Qheader ;;


let read_entry in_channel =
 let offset = Io.r_long in_channel in
 let size = Io.r_long in_channel in
 { offset = offset ; size = size }
;;


let read_vec3 in_channel =
  let x = Io.r_float in_channel in
  let y = Io.r_float in_channel in
  let z = Io.r_float in_channel in
  { x = x ; y = y ; z = z }
;;


let read_bboxshort in_channel =
  let smin_x = Io.r_short in_channel in
  let smin_y = Io.r_short in_channel in
  let smin_z = Io.r_short in_channel in
  let smax_x = Io.r_short in_channel in
  let smax_y = Io.r_short in_channel in
  let smax_z = Io.r_short in_channel in
  { smin = [|smin_x;smin_y;smin_z|] ; smax = [|smax_x;smax_y;smax_z|] }
;;


let read_boundbox in_channel =
  let min = read_vec3 in_channel in
  let max = read_vec3 in_channel in
  { min = min ; max = max }
;;


let read_header in_channel =
  let version = Io.r_long in_channel in
  if version <> 0x1D then
    Printf.fprintf stderr
      "Warning : bad version number. Should be 0x1D but is %d\n" version ;
  let entities_entry = read_entry in_channel in
  let planes_entry = read_entry in_channel in
  let miptex_entry = read_entry in_channel in
  let vertices_entry = read_entry in_channel in
  let visilist_entry = read_entry in_channel in
  let nodes_entry = read_entry in_channel in
  let texinfo_entry = read_entry in_channel in
  let faces_entry = read_entry in_channel in
  let lightmaps_entry = read_entry in_channel in
  let clipnodes_entry = read_entry in_channel in
  let leaves_entry = read_entry in_channel in
  let lfaces_entry = read_entry in_channel in
  let edges_entry = read_entry in_channel in
  let ledges_entry = read_entry in_channel in
  let models_entry = read_entry in_channel in
  { version = version ; entities = entities_entry ; planes = planes_entry ;
    miptex = miptex_entry ; vertices = vertices_entry ;
    visilist = visilist_entry ; nodes = nodes_entry ; texinfo = texinfo_entry ;
    faces = faces_entry ; lightmaps = lightmaps_entry ;
    clipnodes = clipnodes_entry ;leaves = leaves_entry ;
    lfaces = lfaces_entry ; edges = edges_entry ; ledges = ledges_entry ;
    models = models_entry }
;;



let read_entity_line in_channel =
  let nb_read_bytes = ref 0 in
  let entity_desc = ref [] in
  let line = input_line in_channel in
  nb_read_bytes := !nb_read_bytes + String.length line + 1 ;
  if line <> "{" then failwith "Missing {" ;
  let line = ref (input_line in_channel) in
  nb_read_bytes := !nb_read_bytes + String.length !line + 1 ;
  while !line <> "}" do
    let field1_start = String.index_from !line 0 '"' in
    let field1_end = String.index_from !line (field1_start + 1) '"' in
    let field2_start = String.index_from !line (field1_end + 1) '"' in
    let field2_end = String.index_from !line (field2_start + 1) '"' in
    let field1 =
      String.sub !line (field1_start + 1) (field1_end - field1_start - 1) in
    let field2 =
      String.sub !line (field2_start + 1) (field2_end - field2_start - 1) in
    entity_desc := (field1, field2) :: !entity_desc ;
    line := input_line in_channel ;
    nb_read_bytes := !nb_read_bytes + String.length !line + 1
  done ;
  (* On rcupre l'entit inverse, mais a n'est pas grave *)
  (!entity_desc, !nb_read_bytes)
;;



let classify_entity_desc entity_desc =
  try
    begin
    let ent_name = List.assoc "classname" entity_desc in
    match ent_name with
     | "info_player_start" ->
	 let origin = List.assoc "origin" entity_desc in
	 (* x *)
	 let x_string_start = 0 in
	 let x_string_stop = String.index origin ' ' in
	 (* y *)
	 let y_string_start = x_string_stop + 1 in
	 let y_string_stop =
	   String.index_from origin (y_string_start + 1) ' ' in
	 (* z *)
	 let z_string_start = y_string_stop + 1 in
	 let z_string_stop = String.length origin - 1 in
	 let x_string =
	   String.sub origin 0 (x_string_stop - x_string_start) in
	 let y_string = String.sub
	          origin y_string_start (y_string_stop - y_string_start) in
	 let z_string = String.sub
	          origin z_string_start (z_string_stop - z_string_start + 1) in
	 let x = float_of_string x_string in
	 let y = float_of_string y_string in
	 let z = float_of_string z_string in
	 Info_player_start { x = x ; y = y ; z = z }
     | whatever ->
	 Printf.fprintf stderr "Warning unhandled entity type '%s'\n"
	                whatever ;
	 Unhandled_entity whatever
    end
  with Not_found -> Unnamed_entity
;;



let read_entities header in_channel =
  (* Il semblerait que ce soit la taille du chunk - 1. Bizarre... *)
  let size = header.entities.size - 1 in
  seek_in in_channel header.entities.offset ;
  let entities = ref [] in
  let nb_read_bytes = ref 0 in
  (* On lit toutes les entits prsentes *)
  while !nb_read_bytes < size do
    let (entity_desc, nb_read_bytes_for_entity) =
      read_entity_line in_channel in
    (* Ne mmorise les entits que si elles sont intressantes *)
    (match classify_entity_desc entity_desc with
     | Unnamed_entity -> ()
     | Unhandled_entity _ -> ()
     | whatever -> entities := whatever :: !entities) ;
    nb_read_bytes := !nb_read_bytes + nb_read_bytes_for_entity
  done ;
  let tmp = Array.of_list !entities in
  let numentities = Array.length tmp in
  (tmp, numentities)
;;



let read_planes header in_channel =
  let size = header.planes.size in
  let numplanes = size / sizeof_plane_t in
  seek_in in_channel header.planes.offset ;
  let dummy =
    { normal = { x = 0.0 ; y = 0.0 ; z = 0.0 } ; dist = 0.0 ; ptype = 0 } in
  let planes = Array.create numplanes dummy in
  for i = 0 to numplanes - 1 do
    let normal = read_vec3 in_channel in
    let dist = Io.r_float in_channel in
    let ptype = Io.r_long in_channel in
    planes.(i) <-
       { normal = normal ; dist = dist ; ptype = ptype }
    done ;
  (planes, numplanes)
;;


let read_vertices header in_channel =
  let size = header.vertices.size in
  let numvertices = size / sizeof_vertex_t in
  seek_in in_channel header.vertices.offset ;
  let dummy = { vx = 0.0 ; vy = 0.0 ; vz = 0.0 } in
  let vertices = Array.create numvertices dummy in
  for i = 0 to numvertices - 1 do
    let vx = Io.r_float in_channel in
    let vy = Io.r_float in_channel in
    let vz = Io.r_float in_channel in
    vertices.(i) <- { vx = vx ; vy = vy ; vz = vz }
  done ;
  (vertices, numvertices)
;;


let read_visilists header in_channel =
  let size = header.visilist.size in
  let numvisilists = size in
  seek_in in_channel header.visilist.offset ;
  let visilists = String.create numvisilists in
  for i = 0 to numvisilists - 1 do
    let byte = Io.r_byte in_channel in
    visilists.[i] <- Char.chr byte
  done ;
  (visilists, numvisilists)
;;



let read_nodes header in_channel =
  let size = header.nodes.size in
  let numnodes = size / sizeof_node_t in
  seek_in in_channel header.nodes.offset ;
  let dummy = { nplane_id = 0 ; front = 0 ; back = 0 ;
		box = { smin = [||] ; smax = [||] } ;
		nface_id = 0 ; nface_num = 0 } in
  let nodes = Array.create numnodes dummy in
  for i = 0 to numnodes - 1 do
    let nplane_id = Io.r_long in_channel in
    let front = Io.r_short in_channel in
    let back = Io.r_short in_channel in
    let box = read_bboxshort in_channel in
    let nface_id = Io.r_short in_channel in
    let nface_num = Io.r_short in_channel in
    nodes.(i) <- { nplane_id = nplane_id ; front = front ; back = back ;
		box = box ; nface_id = nface_id ; nface_num = nface_num }
    done ;
  (nodes, numnodes)
;;



let read_surfaces header in_channel =
  let size = header.texinfo.size in
  let numsurfaces = size / sizeof_surface_t in
  seek_in in_channel header.texinfo.offset ;
  let dummy = { vectorS = { x = 0.0 ; y = 0.0 ; z = 0.0 } ; distS = 0.0 ;
		vectorT = { x = 0.0 ; y = 0.0 ; z = 0.0 } ; distT = 0.0 ;
		texture_id = 0 ; animated = 0 } in
  let surfaces = Array.create numsurfaces dummy in
  for i = 0 to numsurfaces - 1 do
    let vectorS = read_vec3 in_channel in 
    let distS = Io.r_float in_channel in
    let vectorT = read_vec3 in_channel in 
    let distT = Io.r_float in_channel in
    let texture_id = Io.r_long in_channel in
    let animated = Io.r_long in_channel in
    surfaces.(i) <- { vectorS = vectorS ; distS = distS ; vectorT = vectorT ;
		      distT = distT ; texture_id = texture_id ;
		      animated = animated }
    done ;
  (surfaces, numsurfaces)
;;


let read_faces header in_channel =
  let size = header.faces.size in
  let numfaces = size / sizeof_face_t in
  seek_in in_channel header.faces.offset ;
  let dummy = { plane_id = 0 ; side = 0 ; ledge_id = 0 ; ledge_num = 0 ;
		texinfo_id = 0 ; typelight = 0 ; baselight = 0 ;
	      light1 = 0 ; light2 = 0 ; lightmap = 0 } in
  let faces = Array.create numfaces dummy in
  for i = 0 to numfaces - 1 do
    let plane_id = Io.r_short in_channel in
    let side = Io.r_short in_channel in
    let ledge_id = Io.r_long in_channel in
    let ledge_num = Io.r_short in_channel in
    let texinfo_id = Io.r_short in_channel in
    let typelight = Io.r_byte in_channel in
    let baselight = Io.r_byte in_channel in
    let light1 = Io.r_byte in_channel in
    let light2 = Io.r_byte in_channel in
    let lightmap = Io.r_long in_channel in
    faces.(i) <- { plane_id = plane_id ; side = side ; ledge_id = ledge_id ;
		   ledge_num = ledge_num ; texinfo_id = texinfo_id ;
		   typelight = typelight ; baselight = baselight ;
		   light1 = light1 ; light2 = light2 ; lightmap = lightmap }
    done ;
  (faces, numfaces)
;;


let read_clipnodes header in_channel =
  let size = header.clipnodes.size in
  let numclipnodes = size / sizeof_clipnode_t in
  seek_in in_channel header.clipnodes.offset ;
  let dummy = { planenum = 0 ; cfront = 0 ; cback = 0 } in
  let clipnodes = Array.create numclipnodes dummy in
  for i = 0 to numclipnodes - 1 do
    let planenum = Io.r_long in_channel in
    let cfront = Io.r_short in_channel in
    let cback = Io.r_short in_channel in
    clipnodes.(i) <- { planenum = planenum ; cfront = cfront ; cback = cback }
  done ;
  (clipnodes, numclipnodes)
;;


let read_leaves header in_channel =
  let size = header.leaves.size in
  let numleaves = size / sizeof_leaf_t in
  seek_in in_channel header.leaves.offset ;
  let dummy = { ltype = 0 ; vislist = 0 ;
		lbound = { smin = [||] ; smax = [||] } ;
		lface_id = 0 ; lface_num = 0 ; sndwater = 0 ; sndsky = 0 ;
		sndslimes = 0 ; sndlava = 0 } in
  let leaves = Array.create numleaves dummy in
  for i = 0 to numleaves - 1 do
    let ltype = Io.r_long in_channel in
    let vislist = Io.r_long in_channel in
    let lbound = read_bboxshort in_channel in
    let lface_id = Io.r_short in_channel in
    let lface_num = Io.r_short in_channel in
    let sndwater = Io.r_byte in_channel in
    let sndsky = Io.r_byte in_channel in
    let sndslimes = Io.r_byte in_channel in
    let sndlava = Io.r_byte in_channel in
    leaves.(i) <- { ltype = ltype ; vislist = vislist ;
		    lbound = lbound ; lface_id = lface_id ;
		    lface_num = lface_num ; sndwater = sndwater ;
		    sndsky = sndsky ; sndslimes = sndslimes ;
		    sndlava = sndlava }
  done ;
  (leaves, numleaves)
;;


let read_edges header in_channel =
  let size = header.edges.size in
  let numedges = size / sizeof_edge_t in
  seek_in in_channel header.edges.offset ;
  let dummy = { vertex0 = 0 ; vertex1 = 0 } in
  let edges = Array.create numedges dummy in
  for i = 0 to numedges - 1 do
    let vertex0 = Io.r_short in_channel in
    let vertex1 = Io.r_short in_channel in
    edges.(i) <- { vertex0 = vertex0 ; vertex1 = vertex1 }
  done ;
  (edges, numedges)
;;


let read_models header in_channel =
  let size = header.models.size in
  let nummodels = size / sizeof_model_t in
  seek_in in_channel header.models.offset ;
  let dummy = { bound = { min = { x = 0.0 ; y = 0.0 ; z = 0.0 } ;
			  max = { x = 0.0 ; y = 0.0 ; z = 0.0 } } ;
		origin = { x = 0.0 ; y = 0.0 ; z = 0.0 } ;
		node_id0 = 0 ; node_id1 = 0 ; node_id2 = 0 ; node_id3 = 0 ;
		numleafs = 0 ; face_id = 0 ; face_num = 0 } in
  let models = Array.create nummodels dummy in
  for i = 0 to nummodels - 1 do
    let bound = read_boundbox in_channel in
    let origin = read_vec3 in_channel in
    let node_id0 = Io.r_long in_channel in
    let node_id1 = Io.r_long in_channel in
    let node_id2 = Io.r_long in_channel in
    let node_id3 = Io.r_long in_channel in
    let numleafs = Io.r_long in_channel in
    let face_id = Io.r_long in_channel in
    let face_num = Io.r_long in_channel in
    models.(i) <- { bound = bound ; origin = origin ; node_id0 = node_id0 ;
		    node_id1 = node_id1 ; node_id2 = node_id2 ;
		    node_id3 = node_id3 ; numleafs = numleafs ;
		    face_id = face_id ; face_num = face_num }
  done ;
  (models, nummodels)
;;


let read_lfaces header in_channel =
  let size = header.lfaces.size in
  let numlfaces = size / sizeof_lface in
  seek_in in_channel header.lfaces.offset ;
  let lfaces = Array.create numlfaces 0 in
  for i = 0 to numlfaces - 1 do
    lfaces.(i) <- Io.r_short in_channel
  done ;
  (numlfaces, lfaces)
;;



let read_ledges header in_channel =
  let size = header.ledges.size in
  let numledges = size / sizeof_ledge in
  seek_in in_channel header.ledges.offset ;
  let ledges = Array.create numledges 0 in
  for i = 0 to numledges - 1 do
    ledges.(i) <- Io.r_long in_channel
  done ;
  (ledges, numledges)
;;


let dummy_data = String.make (64*64) ' ' ;;
let make_glide_miptex width height data =
  match (width, height) with
   (* ASPECT_1x1 *)
   | (256, 256) -> (Gr.lod_256, Gr.ASPECT_1x1, data)
   | (128, 128) -> (Gr.lod_128, Gr.ASPECT_1x1, data)
   | (64, 64) -> (Gr.lod_64, Gr.ASPECT_1x1, data)
   | (32, 32) -> (Gr.lod_32, Gr.ASPECT_1x1, data)
   | (16, 16) -> (Gr.lod_16, Gr.ASPECT_1x1, data)
   | (8, 8) -> (Gr.lod_8, Gr.ASPECT_1x1, data)
   | (4, 4) -> (Gr.lod_4, Gr.ASPECT_1x1, data)
   | (2, 2) -> (Gr.lod_2, Gr.ASPECT_1x1, data)
   | (1, 1) -> (Gr.lod_1, Gr.ASPECT_1x1, data)
   (* ASPECT_2x1 *)
   | (256, 128) -> (Gr.lod_256, Gr.ASPECT_2x1, data)
   | (128, 64) -> (Gr.lod_128, Gr.ASPECT_2x1, data)
   | (64, 32) -> (Gr.lod_64, Gr.ASPECT_2x1, data)
   | (32, 16) -> (Gr.lod_32, Gr.ASPECT_2x1, data)
   | (16, 8) -> (Gr.lod_16, Gr.ASPECT_2x1, data)
   | (8, 4) -> (Gr.lod_8, Gr.ASPECT_2x1, data)
   | (4, 2) -> (Gr.lod_4, Gr.ASPECT_2x1, data)
   | (2, 1) -> (Gr.lod_2, Gr.ASPECT_2x1, data)
   (* ASPECT_1x2 *)
   | (128, 256) -> (Gr.lod_256, Gr.ASPECT_1x2, data)
   | (64, 128) -> (Gr.lod_128, Gr.ASPECT_1x2, data)
   | (32, 64) -> (Gr.lod_64, Gr.ASPECT_1x2, data)
   | (16, 32) -> (Gr.lod_32, Gr.ASPECT_1x2, data)
   | (8, 16) -> (Gr.lod_16, Gr.ASPECT_1x2, data)
   | (4, 8) -> (Gr.lod_8, Gr.ASPECT_1x2, data)
   | (2, 4) -> (Gr.lod_4, Gr.ASPECT_1x2, data)
   | (1, 2) -> (Gr.lod_2, Gr.ASPECT_1x2, data)
   (* ASPECT_4x1 *)
   | (256, 64) -> (Gr.lod_256, Gr.ASPECT_4x1, data)
   | (128, 32) -> (Gr.lod_128, Gr.ASPECT_4x1, data)
   | (64, 16) -> (Gr.lod_64, Gr.ASPECT_4x1, data)
   | (32, 8) -> (Gr.lod_32, Gr.ASPECT_4x1, data)
   | (16, 4) -> (Gr.lod_16, Gr.ASPECT_4x1, data)
   | (8, 2) -> (Gr.lod_8, Gr.ASPECT_4x1, data)
   | (4, 1) -> (Gr.lod_4, Gr.ASPECT_4x1, data)
   (* ASPECT_1x4 *)
   | (64, 256) -> (Gr.lod_256, Gr.ASPECT_1x4, data)
   | (32, 128) -> (Gr.lod_128, Gr.ASPECT_1x4, data)
   | (16, 64) -> (Gr.lod_64, Gr.ASPECT_1x4, data)
   | (8, 32) -> (Gr.lod_32, Gr.ASPECT_1x4, data)
   | (4, 16) -> (Gr.lod_16, Gr.ASPECT_1x4, data)
   | (2, 8) -> (Gr.lod_8, Gr.ASPECT_1x4, data)
   | (1, 4) -> (Gr.lod_4, Gr.ASPECT_1x4, data)
   (* ASPECT_8x1 *)
   | (256, 32) -> (Gr.lod_256, Gr.ASPECT_8x1, data)
   | (128, 16) -> (Gr.lod_128, Gr.ASPECT_8x1, data)
   | (64, 8) -> (Gr.lod_64, Gr.ASPECT_8x1, data)
   | (32, 4) -> (Gr.lod_32, Gr.ASPECT_8x1, data)
   | (16, 2) -> (Gr.lod_16, Gr.ASPECT_8x1, data)
   | (8, 1) -> (Gr.lod_8, Gr.ASPECT_8x1, data)
   (* ASPECT_1x8 *)
   | (32, 256) -> (Gr.lod_256, Gr.ASPECT_1x8, data)
   | (16, 128) -> (Gr.lod_128, Gr.ASPECT_1x8, data)
   | (8, 64) -> (Gr.lod_64, Gr.ASPECT_1x8, data)
   | (4, 32) -> (Gr.lod_32, Gr.ASPECT_1x8, data)
   | (2, 16) -> (Gr.lod_16, Gr.ASPECT_1x8, data)
   | (1, 8) -> (Gr.lod_8, Gr.ASPECT_1x8, data)
   (* OTHER CASES *)
   | (_, _) ->
       Printf.fprintf stderr "No texture yet in this format (%d, %d).\n"
	              width height ;
       (Gr.lod_64, Gr.ASPECT_1x1, dummy_data)
;;



(* ******************
Division entire avec arrondi  l'entier suprieur si a ne tombe pas juste.
*)
let div_round_up a b =
  let c = a / b in
  if c * b < a then c + 1 else c
;;

(* ******************
Recherche le LOD de Glide le plus proche d'une dimension de texture.
*)
let round_dimension_to_lod dim =
  if dim > 256 then failwith "round_dimension_to_load : texture to wide"
  else if dim > 128 then Gr.lod_256
  else if dim > 64 then Gr.lod_128
  else if dim > 32 then Gr.lod_64
  else if dim > 16 then Gr.lod_32
  else if dim > 8 then Gr.lod_16
  else if dim > 4 then Gr.lod_8
  else if dim > 2 then Gr.lod_4
  else if dim > 1 then Gr.lod_2
  else if dim > 0 then Gr.lod_1
  else failwith "round_dimension_to_load : texture to small"
;;



let foobar width height data =
  let max_dim = max width height in
  let min_dim = min width height in
  (* On ne sait pas encore si le ratio sera "Gr.ASPECT_ratio:1" *)
  (* ou "Gr.ASPECT_1:ratio". Ce sera dtermin par la dimension *)
  (* la plus grande. Pour le moment, on calcule juste ce ratio. *)
  (* De mme, ce ratio pourra tre invalid vis  vis de Glide.  *)
  (* Par exemple une texture 256x2 aura moralement un ratio de  *)
  (* la forme "Gr.ASPECT_128:1" ce qui est invalide. Il faudra  *)
  (* dans ce cas faire quelque chose.                           *)
  let ratio = div_round_up max_dim min_dim in
  let aspect =
    if width > height then
      begin
      (* Forme 1:x *)
      match ratio with
       | 1 -> Gr.ASPECT_1x1
       | 2 -> Gr.ASPECT_1x2
       | 4 -> Gr.ASPECT_1x4
       | 8 -> Gr.ASPECT_1x8
       | _ -> failwith "foobar : invalid ratio"
      end
    else
      begin
      (* Forme x:1 *)
      match ratio with
       | 1 -> Gr.ASPECT_1x1
       | 2 -> Gr.ASPECT_2x1
       | 4 -> Gr.ASPECT_4x1
       | 8 -> Gr.ASPECT_8x1
       | _ -> failwith "foobar : invalid ratio"
      end in
  let lod = round_dimension_to_lod max_dim in
  (lod, aspect, data)
;;



(* Retourne directement un tex_info Glide infr  partie des donnes de *)
(* textures Quake. *)
let read_miptexs header in_channel =
  seek_in in_channel header.miptex.offset ;
  let nummiptex = Io.r_long in_channel in
  (* Initialize la table des offsets *)
  let offset = Array.create nummiptex 0 in
  for i = 0 to nummiptex - 1 do offset.(i) <- Io.r_long in_channel done ;
  let dummy = { Gr.small_lod = Gr.lod_1 ; Gr.large_lod = Gr.lod_1 ;
		Gr.aspect_ratio = Gr.ASPECT_1x1 ; Gr.format = Gr.TEXFMT_P_8 ;
		Gr.data = "" } in
  let miptexs = Array.create nummiptex dummy in
  let texture_scaling = Array.create nummiptex 0.0 in
  for i = 0 to nummiptex - 1 do
    Printf.fprintf stderr "Texture %d ...\n" i ;
    (* Aller  l'offset de la texture *)
    let texture_offset = header.miptex.offset + offset.(i) in
    seek_in in_channel texture_offset ;
    let name = Io.r_string_fixed 16 in_channel in
    let width = Io.r_long in_channel in
    let height = Io.r_long in_channel in
    let offset1 = Io.r_long in_channel in
    let offset2 = Io.r_long in_channel in
    let offset4 = Io.r_long in_channel in
    let offset8 = Io.r_long in_channel in
    seek_in in_channel (texture_offset + offset1) ;
    let data = Io.r_string_fixed (width * height) in_channel in
    let (lod, ratio, texels) = make_glide_miptex width height data in
    miptexs.(i) <- { Gr.small_lod = lod ; Gr.large_lod = lod ;
		     Gr.aspect_ratio = ratio ;
		     Gr.format = Gr.TEXFMT_P_8 ; Gr.data = texels } ;
    (* Bugg, mais c'est une approximation pour le moment *)
    (* Calcule le facteur permettant de passer en mode de coordonnes *)
    (*  la Glide. En effet, sous Glide, les extends vont tous de 0  *)
    (* 256 pour la plus grande dimension de la texture. Alors que     *)
    (* Quake donne ses coordonnes relativement  la taille de sa     *)
    (* texture (c'est  dire qu'un pixel a une taille de 1 quelque    *)
    (* soit la dimension de la texture).                              *)
    let scale = 256.0 /. (float height) in
    texture_scaling.(i) <- scale
    done ;
  (miptexs, texture_scaling, nummiptex)
;;


let load_file filename =
  Printf.fprintf stderr "Loading file '%s'.\n" filename ;
  let in_channel = open_in_bin filename in
  let bsp_header = read_header in_channel in
  let (entities, numentities) = read_entities bsp_header in_channel in
  let (planes, numplanes) = read_planes bsp_header in_channel in
  let (vertices, numvertices) = read_vertices bsp_header in_channel in
  let (visilists, numvisilists) = read_visilists bsp_header in_channel in
  let (nodes, numnodes) = read_nodes bsp_header in_channel in
  let (faces, numfaces) = read_faces bsp_header in_channel in
  let (clipnodes, numclipnodes) = read_clipnodes bsp_header in_channel in
  let (leaves, numleaves) = read_leaves bsp_header in_channel in
  let (numlfaces, lfaces) = read_lfaces bsp_header in_channel in
  let (edges, numedges) = read_edges bsp_header in_channel in
  let (ledges, numledges) = read_ledges bsp_header in_channel in
  let (models, nummodels) = read_models bsp_header in_channel in
  let (surfaces, numsurfaces) = read_surfaces bsp_header in_channel in
  let (miptexs, texture_scaling, nummiptexs) =
    read_miptexs bsp_header in_channel in
  Global.numentities := numentities ;
  Global.entities := entities ;
  Global.nummodels := nummodels ;
  Global.models := models ;
  Global.numnodes := numnodes ;
  Global.nodes := nodes ;
  Global.numplanes := numplanes ;
  Global.planes := planes ;
  Global.numleaves := numleaves ;
  Global.leaves := leaves ;
  Global.numvisilists := numvisilists ;
  Global.visilists := visilists ;
  Global.numlfaces := numlfaces ;
  Global.lfaces := lfaces ;
  Global.numfaces := numfaces ;
  Global.faces := faces ;
  Global.numclipnodes := numclipnodes ;
  Global.clipnodes := clipnodes ;
  Global.numledges := numledges ;
  Global.ledges := ledges ;
  Global.numedges := numedges ;
  Global.edges := edges ;
  Global.numvertices := numvertices ;
  Global.vertices := vertices ;
  Global.numsurfaces := numsurfaces ;
  Global.surfaces := surfaces ;
  Global.nummiptexs := nummiptexs ;
  Global.miptexs := miptexs ;
  Global.texture_scaling := texture_scaling ;
  (* C'est juste pour patcher le fait que quand on decode, on ne verifie *)
  (* pas si une partie de l'octet est inutilisee, donc ca risque de *)
  (* nous faire taper en dehors de ce tableau. Pour regler ca, on alloue *)
  (* directement 8 feuilles de plus dans le tableaux de flags visibles pour *)
  (* les feuilles. *)
  let vis_leaves_size = numleaves + 8 in
  Global.visible_leaves := Array.create vis_leaves_size false ;
  Global.visible_nodes := Array.create numnodes false ;
  Global.visible_faces := Array.create numfaces false ;
  flush stderr
;;
