(* ********************** *)
(* Basic type definitions *)
(* ********************** *)

type vec3_t = {
  mutable x : float ;   (* Horizontal *)
  mutable y : float ;   (* Horizontal *)
  mutable z : float     (* Vertical *)
  } ;;

type scalar_t = float ;;    (* 32 bit float *)

type long_t = int ;;        (* 32 bit signed integer *)
type short_t = int ;;       (* 16 bit signed integer *)

type ulong_t = int ;;       (* 32 bit unsigned integer *)
type ushort_t = int ;;      (* 16 bit unsigned integer *)
type uchar_t = int ;;       (* 8 bit unsigned integer *)


(* ************** *)
(* Bounding boxes *)
(* ************** *)

type boundbox_t = {
  min : vec3_t ;     (* minimum values of X, Y, Z *)
  max : vec3_t       (* maximum values of X, Y, Z *)
  } ;;

type bboxshort_t = {
  smin : short_t array ;       (* minimum values of X, Y, Z *)
  smax : short_t array         (* maximum values of X, Y, Z *)
  } ;;


(* *************** *)
(* BSP file header *)
(* *************** *)

type dentry_t = {
  offset : long_t ;
  size : long_t
  } ;;

type dheader_t = {
  version : long_t ;     (* Model version. Must be 0x17 (23) *)
  entities : dentry_t ;  (* List of entities *)
  planes : dentry_t ;    (* Map planes. num=size/sizeof(plane_t) *)
  miptex : dentry_t ;    (* Wall textures *)
  vertices : dentry_t ;  (* Map vertices. num=size/sizeof(vertex_t) *)
  visilist : dentry_t ;  (* Leaves visibility lists *)
  nodes : dentry_t ;     (* BSP nodes. num=size/sizeof(node_t) *)
  texinfo : dentry_t ;   (* Texture for faces. num=size/sizeof(texinfo_t) *)
  faces : dentry_t ;     (* Faces of each surface. num=size/sizeof(face_t) *)
  lightmaps : dentry_t ; (* Wall light maps *)
  clipnodes : dentry_t ; (* Clip nodes of model. num=size/sizeof(clipnode_t) *)
  leaves : dentry_t ;    (* BSP leaves. num=size/sizeof(leaf_t) *)
  lfaces : dentry_t ;    (* List of faces *)
  edges : dentry_t ;     (* Edges of faces. num=size/sizeof(edge_t) *)
  ledges : dentry_t ;    (* List of edges *)
  models : dentry_t      (* nummodels = size / sizeof (model_t) *)
  } ;;


(* ****** *)
(* Models *)
(* ****** *)
type model_t = {
  bound : boundbox_t ;      (* The bounding box of the model *)
  origin : vec3_t ;         (* Origin of the model, usually (0, 0, 0) *)
  node_id0 : long_t ;       (* Index of first BSP node *)
  node_id1 : long_t ;       (* Index of the first clip node *)
  node_id2 : long_t ;       (* Index of the second clip node *)
  node_id3 : long_t ;       (* Usually 0 *)
  numleafs : long_t ;       (* Number of BSP leaves *)
  face_id : long_t ;        (* Index of faces of this model *)
  face_num : long_t         (* Number of faces of this model *)
  } ;;


(* ******** *)
(* Vertices *)
(* ******** *)

type vertex_t = {
  vx : float ;        (* X, Y, Z coordinates of the vertex *)
  vy : float ;        (* usually some integer value *)
  vz : float          (* but coded in floating point *)
  } ;;


(* ***** *)
(* Edges *)
(* ***** *)

type edge_t = {
  vertex0 : ushort_t ;   (* Index of the start vertex. [0, numvertices[ *)
  vertex1 : ushort_t     (* Index of the end vertex. [0, numvertices[ *)
  } ;;


(* ***** *)
(* Faces *)
(* ***** *)

type face_t = {
  plane_id : ushort_t ; (* Plane in which the face lies. [0, numplanes[ *)
  side : ushort_t ;     (* 0 if in front of the plane, 1 if behind the plane *)
  ledge_id : long_t ;   (* First edge in the list of edges. [0, numledges[ *)
  ledge_num : ushort_t ;   (* Number of edges in the list of edges *)
  texinfo_id : ushort_t ;  (* Index of the texture info the face is part of *)
                           (* [0, numtexinfos[ *)
  typelight : uchar_t ;    (* Type of lighting for the face *)
  baselight : uchar_t ;    (* From 0xFF (dark) to 0 (bright) *)
  light1 : uchar_t ;       (* Two additionnal ... *)
  light2 : uchar_t ;       (* ... light models *)
  lightmap : long_t   (* Pointer inside the general light map, or -1 ... *)
                      (* ... This defines the start of the face light map *)
  } ;;


(* ************** *)
(* BSP tree nodes *)
(* ************** *)

type node_t = {
  nplane_id : long_t ;   (* The plane that splits the node. [0, numplqnes[ *)
  front : ushort_t ;     (* If bit15 = 0, index of front child node *)
                         (* If bit15 = 1, ~front = index of child leaf *)
  back : ushort_t ;      (* If bit15 = 0, index of back child node *)
                         (* If bit15 = 1, ~back = index of child leaf *)
  box : bboxshort_t ;    (* Bounding box of node and its children *)
  nface_id : ushort_t ;  (* Index of the first polygon in the node *)
  nface_num : ushort_t   (* Number of faces in the node *)
  } ;;


(* *************** *)
(* BSP tree leaves *)
(* *************** *)
type dleaf_t = {
  ltype : long_t ;      (* Special type of leaf *)
                        (* -1 : ordinary leaf *)
                        (* -2 : the leaf is entierely inside a solid *)
                        (* -3 : water, the vision is troubled *)
                        (* -4 : Slime, green acid taht hurst the player *)
                        (* -5 : Lava, vision turns red and player burns *)
                        (* -6 : Behaves like water, but is used for sky *)
                        (* Others seem to be water (not yet implemented ? *)
  vislist : long_t ;    (* Beginning of the visibility lists. Must be -1 ... *)
                        (* of in [0, numvislist[ *)
  lbound : bboxshort_t ;  (* Bounding box of the leaf *)
  lface_id : ushort_t ;   (* First item of the list of faces. [0, numfaces[ *)
  lface_num : ushort_t ;  (* Number of faces in the leaf *)
  sndwater : uchar_t ;    (* Level of the 4 ambient sounds ... *)
  sndsky : uchar_t ;      (* ... 0 is no sound ... *)
  sndslimes : uchar_t ;   (* ... 0xFF is maximim volume *)
  sndlava : uchar_t
  } ;;


(* ****** *)
(* Planes *)
(* ****** *)

type plane_t = {
  normal : vec3_t ;   (* Vector orthogonal to plane (Nx, Ny, Nz) with ... *)
                      (* Nx2+Ny2+Nz2 = 1 *)
  dist : scalar_t ;   (* Offset to plane along the normal vector ... *)
                      (* ... Distance from (0, 0, 0) to the plane *)
  ptype : long_t      (* Type of plane, depending on normal vector *)
                      (* 0 : Axial plane, in X *)
                      (* 1 : Axial plane, in Y *)
                      (* 2 : Axial plane, in Z *)
		      (* 3 : Non axial plane, roughlt toward X *)
		      (* 4 : Non axial plane, roughlt toward Y *)
		      (* 5 : Non axial plane, roughlt toward Z *)
  } ;;


(* ********** *)
(* Clip nodes *)
(* ********** *)

type clipnode_t = {
  planenum : ulong_t ;  (* The plane which splits the node *)
  cfront : short_t ;    (* If positive, id of front child node *)
                        (* If -2, the front part is inside the model *)
                        (* If -1, the front part is outside the model *)
  cback : short_t       (* If positive, id of back child node *)
                        (* If -2, the back part is inside the model *)
                        (* If -1, the back part is outside the model *)
  } ;;


(* **************** *)
(* Texture surfaces *)
(* **************** *)

type surface_t = {
  vectorS : vec3_t ;       (* S vector, horizontal in texture space *)
  distS : scalar_t ;       (* Horizontal offset in texture space *)
  vectorT : vec3_t ;       (* T vector, vertical in texture space *)
  distT : scalar_t ;       (* Vertical offset in texture space *)
  texture_id : ulong_t ;   (* Index of mip texture. [0, numtex[ *)
  animated : ulong_t       (* 0 for ordinary textures, 1 for water *)
  } ;;


(* *********************************************************** *)
(* Size of structures in C. Used to compute number of elements *)
(* *********************************************************** *)

let sizeof_plane_t = 20 ;;
let sizeof_vertex_t = 12 ;;
let sizeof_node_t = 24 ;;
let sizeof_face_t =  20 ;;
let sizeof_clipnode_t = 8 ;;
let sizeof_leaf_t = 28 ;;
let sizeof_model_t = 64 ;;
let sizeof_edge_t = 4 ;;
let sizeof_surface_t = 40 ;;

(* For lists of faces and edges *)
let sizeof_lface = 2 ;;
let sizeof_ledge = 4 ;;



(* ************************ *)
(* Custom type for entities *)
(* ************************ *)

type entity_t =
  | Info_player_start of vec3_t
  | Unhandled_entity of string
  | Unnamed_entity
;;
